import { rectDistance } from "./rect-distance";
import { createRandom } from "./helpers";
import { polygonBounds, polygonScale, polygonTranslate } from "geometric";
import Quadtree from "@timohausmann/quadtree-js";

let baseLayout = [];

onmessage = (e) => {
  let {
    elementWidth,
    elementHeight,
    layoutWidth,
    layoutHeight,
    cellDimensions,
    contentRects,
    minContentGap,
    minPatternGap,
    seed,
    fillIterations,
    viewboxMode,
  } = e.data;

  if (!viewboxMode) {
    layoutWidth = elementWidth;
    layoutHeight = elementHeight;
    baseLayout = [];
  }

  const layoutTree = new Quadtree({
    x: 0,
    y: 0,
    width: layoutWidth,
    height: layoutHeight,
  });

  const contentTree = new Quadtree({
    x: 0,
    y: 0,
    width: layoutWidth,
    height: layoutHeight,
  });

  contentRects.forEach((r) => contentTree.insert(r));

  if (!viewboxMode || baseLayout.length === 0) {
    const random = createRandom(seed);

    const isFixedAspectRatio = !cellDimensions.width && !cellDimensions.height;

    for (let i = 0; i < fillIterations; i++) {
      const rect = {};

      if (isFixedAspectRatio) {
        const rectSize = random(cellDimensions.min, cellDimensions.max);
        const x = random(0, layoutWidth - rectSize);
        const y = random(0, layoutHeight - rectSize);

        rect.x = x;
        rect.y = y;
        rect.width = rectSize;
        rect.height = rectSize;
      } else {
        const rectWidth = random(
          cellDimensions.width.min,
          cellDimensions.width.max
        );
        const rectHeight = random(
          cellDimensions.height.min,
          cellDimensions.height.max
        );
        const x = random(0, layoutWidth - rectWidth);
        const y = random(0, layoutHeight - rectHeight);

        rect.x = x;
        rect.y = y;
        rect.width = rectWidth;
        rect.height = rectHeight;
      }

      rect.vertices = [
        [rect.x, rect.y],
        [rect.x + rect.width, rect.y],
        [rect.x + rect.width, rect.y + rect.height],
        [rect.x, rect.y + rect.height],
      ];

      rect.id = random(0, 100_000, 1);

      if (rectDistance(rect, layoutTree.retrieve(rect)) >= minPatternGap) {
        baseLayout.push(rect);
        layoutTree.insert(rect);
      }
    }
  }

  const horizontalRatio = elementWidth / layoutWidth;
  const verticalRatio = elementHeight / layoutHeight;

  const ratio = Math.max(horizontalRatio, verticalRatio);

  const centerShiftX = (elementWidth - layoutWidth * ratio) / 2;
  const centerShiftY = (elementHeight - layoutHeight * ratio) / 2;

  const container = {
    x: 0,
    y: 0,
    width: elementWidth,
    height: elementHeight,
  };

  const rects = [];

  let i = baseLayout.length;
  while (i--) {
    const current = baseLayout[i];

    if (rectDistance(current, rects) >= minPatternGap) {
      rects.push(current);
    }
  }

  postMessage({
    rects: rects
      .map((rect) => {
        if (viewboxMode) {
          const scaled = polygonScale(rect.vertices, ratio, [0, 0]);
          const translatedX = polygonTranslate(scaled, 0, centerShiftX);
          const translatedY = polygonTranslate(translatedX, 90, centerShiftY);
          const bounds = polygonBounds(translatedY);

          return {
            x: bounds[0][0],
            y: bounds[0][1],
            width: bounds[1][0] - bounds[0][0],
            height: bounds[1][1] - bounds[0][1],
            id: rect.id,
          };
        } else {
          return rect;
        }
      })
      .filter(
        (rect) =>
          rectDistance(rect, contentTree.retrieve(rect)) >= minContentGap &&
          rectDistance(rect, [container]) === 0
      ),
  });
};
